import java.io.*;
import java.util.*;
import java.util.regex.*;

/**
 * Representation of a MEAPsoft edit decision list file.
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */

public class EDLFile extends MEAPFile
{
    //Vector chunks;  // Vector of EDLChunks
    // keep them sorted
    MaxHeap chunks;  
    
    public  EDLFile(String fn)
    {
        filename = fn;

        //chunks = new Vector();
        chunks = new MaxHeap();
    }

    /**
     * Read in an EDL file
     */
    public void readFile() throws IOException, ParserException
    {
        BufferedReader in  = new BufferedReader(new FileReader(filename));

        String audioFile="";
        double dstTime;
        double chunkLength;
        double srcTime;

        // each line (excluding comments) should look like: 
        // dest_time src_filename src_time chunk_length cmd1 cmd2 ... 
        Pattern commentPattern = Pattern.compile("#\\.*");
        Pattern linePattern = Pattern.compile("\\s*([^#\\s]+)\\s*");

        // Parse each line of the input file            
        long lineno = 0;
        String line;
        while((line = in.readLine()) != null) 
        { 
            lineno++;
            
            // extract any comments from the current line
            String comment = "";
            Matcher c = commentPattern.matcher(line+"\n");
            if(c.find())
            {
                // comments go all the way to the end of the line
                comment = c.group() + line.substring(c.end()) + "\n";
                line = line.substring(0, c.start());
            }

            Matcher p = linePattern.matcher(line);
            // is there anything else?
            if(!p.find())
                continue;
            try { dstTime = Double.parseDouble(p.group(1)); }
            catch(NumberFormatException nfe) { 
                throw new ParserException(filename, lineno, 
                                          "Could not parse dest start time \"" 
                                          + p.group(1)  + "\".");  }

            if(!p.find())
                throw new ParserException(filename, lineno, 
                                          "Could not find source filename.");
            audioFile = p.group(1);

            if(!p.find())
                throw new ParserException(filename, lineno, 
                                          "Could not find source start time.");
            try { srcTime = Double.parseDouble(p.group(1)); }
            catch(NumberFormatException nfe) { 
                throw new ParserException(filename, lineno, 
                                          "Could not parse source start time \"" 
                                          + p.group(1)  + "\".");  }

            if(!p.find())
                throw new ParserException(filename, lineno, 
                                          "Could not find chunk length.");
            try { chunkLength = Double.parseDouble(p.group(1)); }
            catch(NumberFormatException nfe) { 
                throw new ParserException(filename, lineno, 
                                          "Could not parse chunk length \"" 
                                          + p.group(1)  + "\".");  }
            
            EDLChunk ch = new EDLChunk(audioFile, srcTime, chunkLength, dstTime);
            ch.comment = comment;

            // everything else on the current line is a command
            while(p.find())
                ch.commands.add(p.group(1));
            
            chunks.add(ch);
        }
        
        in.close();
    }
    
    /**
     * Write the contents of this FeatFile
     */
    protected void write(Writer w) throws IOException
    {
        // write the header
        w.write("# dst_time src_filename src_onset_time src_chunk_length [commands]\n");

        // its much easier to read if the output is in sorted in time;
        chunks.sort();
        
        Iterator i = chunks.iterator();
        while(i.hasNext())
            w.write(i.next().toString());
    }
}

