import javax.sound.sampled.*;
import java.io.*;
import java.util.*;
import java.util.regex.*;
import gnu.getopt.*;

import featextractors.*;

/**
 * Program that extracts features from the chunks listed in the input
 * files. 
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */
public class FeatExtractor extends MEAPUtil
{
    // Files to use
    String[] featFiles;
    String outFile = null;
    
    // all of our feature extractors...
    Vector featExts;
    String feat_names = "";

    public void printUsageAndExit() 
    {
        System.out.println("Usage: FeatExtractor [-options] file1.feat file2.feat ... \n\n" + 
             "  where options include:\n" + 
             "    -o output_file  append features into output file (defaults to input file)");
        printCommandLineOptions('f');
        System.out.println();
        System.exit(0);
    }

    public FeatExtractor(String[] args) 
    {
        if(args.length == 0)
            printUsageAndExit();
        
        // Parse arguments
        String argString = "f:o:";
        featExts = parseFeatureExtractor(args, argString);

        Getopt opt = new Getopt("FeatExtracter", args, argString);
        opt.setOpterr(false);
        
        int c = -1;
        while ((c = opt.getopt()) != -1) 
        {
            switch(c) 
            {
            case 'o':
                outFile = opt.getOptarg();
                break;
            case 'f': // already handled above
                break;
            case '?':
                printUsageAndExit();
                break;
            default:
                System.out.print("getopt() returned " + c + "\n");
            }
        }
        
        // parse arguments
        int ind = opt.getOptind();
        if(ind > args.length)
            printUsageAndExit();

        featFiles = new String[args.length - ind];
        for(int i=ind; i<args.length; i++)
            featFiles[i-ind] = args[i];
        
        // What are the feature names?
        for(int i = 0; i < featExts.size(); i++)
            feat_names += featExts.get(i).getClass().getName() + " ";
        //chop off last space
        feat_names = feat_names.substring(0, feat_names.length()-1);
    }
    
    private FeatFile processFeatFile(String filename) throws IOException, ParserException
    {
        STFT stft = null;
        FeatFile file = new FeatFile(filename);
        file.readFile();
        
        file.featureDescription = "";
        boolean wroteFeatDesc = false;
        
        String lastAudioFile="";
        Iterator c = file.chunks.iterator();
        while(c.hasNext())
        {
            FeatChunk ch = (FeatChunk)c.next();
            // lets get some new features
            // compute the STFT of the whole file if this is the first 
            // time we have seen it
            if(!ch.srcFile.equals(lastAudioFile))
            {
                AudioInputStream ais = openInputStream(ch.srcFile);

                // store STFT of the whole file
                int histSize = (int)(ais.getFrameLength()/frameSize); 
                stft = new STFT(ais, frameSize, histSize);
                stft.start();
                ais.close();
            }
            lastAudioFile = ch.srcFile;

            // compute features from the STFT
            Iterator i = featExts.listIterator();
            while(i.hasNext())
            {
                FeatureExtractor fe = (FeatureExtractor)i.next();
                double[] feats = fe.features(stft, stft.seconds2fr(ch.startTime), 
                                             (int)stft.seconds2fr(ch.length));
                ch.addFeature(feats);

                // what features are we adding?  
                if(!wroteFeatDesc)
                    file.featureDescription += fe.getClass().getName() + "(" + feats.length + ") ";
            }

            wroteFeatDesc = true;
        }

        return file;
    }
    

    public void go() 
    {
        // process supplied files
        for(int i = 0; i < featFiles.length; i++) 
        {
            String fn = outFile;
            if(fn == null)
                fn = featFiles[i];
            System.out.println("Extracting features (" + feat_names + 
                               ") from " + featFiles[i] + " to " + fn + "."); 
            
            long startTime = System.currentTimeMillis();
            try
            {
                FeatFile f = processFeatFile(featFiles[i]);
                f.writeFile(outFile);
            }
            catch(Exception e) 
            {
                e.printStackTrace();
                System.exit(1);
            }
            System.out.println("Done.  Took " + 
                               ((System.currentTimeMillis() - startTime)/1000.0)
                               + "s");
        }
    }
    
    
    public static void main(String[] args) 
    {
        FeatExtractor o2or = new FeatExtractor(args);
        o2or.go();
        System.exit(0);
    }
}

