import java.io.*;
import java.util.*;
import java.util.regex.*;

/**
 * Representation of a MEAPsoft segment/feature file.
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */

public class FeatFile extends MEAPFile
{
    String featureDescription = "";
    Vector chunks;  // Vector of FeatChunks
    
    public FeatFile(String fn)
    {
        filename = fn;

        chunks = new Vector(100, 0);
    }

    /**
     * Read in an feature file
     */
    public void readFile() throws IOException, ParserException
    {
        BufferedReader in  = new BufferedReader(new FileReader(filename));

        String audioFile;
        double chunkStartTime;
        double chunkLength;

        // each line (excluding comments) should look like: 
        // audioFile chunkStartTime chunkLength feature1 feature2 ...
        Pattern commentPattern = Pattern.compile("#\\.*");
        Pattern linePattern = Pattern.compile("\\s*([^#\\s]+)\\s*");

        // Parse each line of the input file            
        String header = "# Features: ";
        boolean haveWrittenHeader = false;
        long lineno = 0;
        String line;
        while((line = in.readLine()) != null) 
        { 
            lineno++;
            
            // extract any comments from the current line
            String comment = "";
            Matcher c = commentPattern.matcher(line+"\n");
            if(c.find())
            {
                // comments go all the way to the end of the line
                comment = c.group() + line.substring(c.end()) + "\n";
                line = line.substring(0, c.start());
            }

            Matcher p = linePattern.matcher(line);
            // is there anything else?
            if(!p.find())
                continue;
            audioFile = p.group(1);
            if(!p.find())
                throw new ParserException(filename, lineno, 
                                          "Could not find chunk start time.");
            try { chunkStartTime = Double.parseDouble(p.group(1)); }
            catch(NumberFormatException nfe) { 
                throw new ParserException(filename, lineno, 
                                          "Could not parse chunk start time \"" 
                                          + p.group(1)  + "\".");  }
            if(!p.find())
                throw new ParserException(filename, lineno, 
                                          "Could not find chunk length.");
            try { chunkLength = Double.parseDouble(p.group(1)); }
            catch(NumberFormatException nfe) { 
                throw new ParserException(filename, lineno, 
                                          "Could not parse chunk length \"" 
                                          + p.group(1)  + "\".");  }
            
            FeatChunk ch = new FeatChunk(audioFile, chunkStartTime, chunkLength);
            ch.comment = comment;

            // save the remaining features on the line
            while(p.find())
            {
                // what kind of feature is this?  If its not a double then its a string;
                try 
                {
                    ch.addFeature(Double.parseDouble(p.group(1)));
                }
                catch (NumberFormatException e)
                {
                    ch.addFeature(p.group(1));
                }
            }

            chunks.add(ch);

            if(comment.startsWith("# Features: "))
            {
                featureDescription = comment.substring(12);
            }
        }

        in.close();
    }
    
    /**
     * Write the contents of this FeatFile
     */
    protected void write(Writer w) throws IOException
    {
        // write the header
        w.write("# filename onset_time chunk_length [features]\n# Features: ");
        w.write(featureDescription + "\n");

        Iterator i = chunks.iterator();
        while(i.hasNext())
            w.write(i.next().toString());
    }
}
