import javax.sound.sampled.*;
import java.io.*;
import java.util.*;
import gnu.getopt.*;
import java.text.NumberFormat;

import featextractors.*;

/**
 * Program that composes a mashup by replacing each segment of the
 * input file with the chunk in a given chunk database whos features
 * most closely match those of the input chunk.  
 *
 * Produces an EDL from two feature files.  Basically an offline
 * version of Mike's mashup generator.
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */
public class MashupComposer extends MEAPUtil
{
    String outFileName = "mashup.edl";
    
    EDLFile outFile;
    FeatFile dstFile;
    FeatFile DBFile;
    int[] featdim = null;
    ChunkDist dist;

    public void printUsageAndExit() 
    {
        System.out.println("Usage: MashupComposer [-options] dest.feat chunkdb.feat \n\n" + 
		       "  where options include:\n" + 
		       "    -o output_file  the file to write the output to (defaults to mashup.edl)");
        printCommandLineOptions('i');
        printCommandLineOptions('d');
        System.out.println();
        System.exit(0);
    }

    public MashupComposer(String featFN, String DBFN, String outFN, ChunkDist cd)
    {
        dstFile = new FeatFile(featFN);
        DBFile = new FeatFile(DBFN);
        outFile = new EDLFile(outFN);
        dist = cd;
        featdim = cd.featdim;
    }

    public MashupComposer(String[] args) 
    {
        if(args.length == 0)
            printUsageAndExit();

        // Parse arguments
        String argString = "d:i:o:";
        featdim = parseFeatDim(args, argString);
        dist = parseChunkDist(args, argString, featdim);
        
        Getopt opt = new Getopt("MashupComposer", args, argString);
        opt.setOpterr(false);

        int c = -1;
        while ((c = opt.getopt()) != -1) 
        {
            switch(c) 
            {
            case 'o':
                outFileName = opt.getOptarg();
                break;
            case 'd':  // already handled above
                break;
            case 'i':  // already handled above
                break;
            case '?':
                printUsageAndExit();
                break;
            default:
                System.out.print("getopt() returned " + c + "\n");
            }
        }

        // parse arguments
        int ind = opt.getOptind();
        if(ind > args.length)
            printUsageAndExit();
        
        dstFile = new FeatFile(args[args.length-2]);
        DBFile = new FeatFile(args[args.length-1]);
        outFile = new EDLFile(outFileName);

        System.out.println("Composing " + outFileName + 
                           " from " +  args[args.length-2] +
                           " (chunk db: " + args[args.length-1] + ").");
    }

    public EDLFile compose()
    {
        NumberFormat fmt = NumberFormat.getInstance();
        fmt.setMaximumFractionDigits(3);

        try
        {
            dstFile.readFile();
            DBFile.readFile();
        }
        catch(Exception e)      
        {
            e.printStackTrace();
            System.exit(1);
        } 

        Iterator dstchunks = dstFile.chunks.iterator();
        while(dstchunks.hasNext())
        {
            FeatChunk currChunk = (FeatChunk)dstchunks.next();
            double mindist = Double.MAX_VALUE;
            FeatChunk match = null;
            
            // find closest match to currChunk in DB
            Iterator i = DBFile.chunks.iterator();
            while(i.hasNext())
            {
                FeatChunk c = (FeatChunk)i.next();
                double d = dist.distance(currChunk, c);

                if(d < mindist)
                {
                    mindist = d;
                    match = c;
                }
            }
            
            // turn match chunk into an EDL chunk 
            EDLChunk nc = new EDLChunk(match, currChunk.startTime);
            nc.comment = "    # dist = " + fmt.format(mindist);
            outFile.chunks.add(nc);
        }

        return outFile;
    } 

    public void go()
    {
        compose();

        try
        {
            outFile.writeFile();
        }
        catch(IOException e)
        {
            e.printStackTrace();
            System.exit(1);
        }
    }

    public static void main(String[] args) 
    {
        MashupComposer m = new MashupComposer(args);
        long startTime = System.currentTimeMillis();
        m.go();
        System.out.println("Done. Took " +
                           ((System.currentTimeMillis() - startTime)/1000.0)
                           + "s");
        System.exit(0);
    }
}
