import javax.sound.sampled.*;
import java.io.*;
import java.util.*;
import gnu.getopt.*;
import java.text.NumberFormat;

import featextractors.*;

/**
 * Program that composes an EDL by hopping around from chunk to chunk.
 * Starts from a the first chunk in the source file.  Always picks the
 * nearest neighbor in feature space from current chunk.  Should do
 * something similar to Dan Ellis' ordercols.m.
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */
public class NNComposer extends SortComposer
{
    public void printUsageAndExit() 
    {
        System.out.println("Usage: NNComposer [-options] features.feat \n\n" + 
		       "  where options include:\n" + 
		       "    -o output_file  the file to write the output to (defaults to sorted.edl)");
        printCommandLineOptions('i');
        printCommandLineOptions('d');
        System.out.println();
        System.exit(0);
    }

    public NNComposer(String featFN, String outFN, ChunkDist cd)
    {
        super(featFN, outFN, cd);
    }

    public NNComposer(String[] args) 
    {
        super(args);
    }

    public EDLFile compose()
    {
        try
        {
            featFile.readFile();
        }
        catch(Exception e)      
        {
            e.printStackTrace();
            System.exit(1);
        } 

        // initial chunk - pick it at random:
        //int randIdx = (int)Math.floor(featFile.chunks.size()*Math.random());
        // start with the first chunk
        FeatChunk currChunk = (FeatChunk)featFile.chunks.get(0);
        
        dist.setTarget(currChunk);
        
        // maintain a set of chunks sorted using dist from targetChunk
        MinHeap chunks = new MinHeap(dist);
        chunks.addAll(featFile.chunks);
        
        NumberFormat fmt = NumberFormat.getInstance();
        fmt.setMaximumFractionDigits(3);
        
        double currTime = 0;
        while(chunks.size() > 0)
        {
            dist.setTarget(currChunk);
            chunks.rebuildHeap();
            
            currChunk = (FeatChunk)chunks.deleteMin();
            
            // turn currChunk into an EDL chunk 
            EDLChunk nc = new EDLChunk(currChunk, currTime);

            if(debug)
            {
                nc.comment = "    # feats = ";
                double[] feat = currChunk.getFeatures(featdim);
                for(int x = 0; x < feat.length-1; x++)
                    nc.comment += fmt.format(feat[x]) + ", ";
                nc.comment += fmt.format(feat[feat.length-1]);
            }

            outFile.chunks.add(nc);
            
            currTime += currChunk.length;
        }
        
        return outFile;
    } 
    
    public static void main(String[] args) 
    {
        NNComposer m = new NNComposer(args);
        long startTime = System.currentTimeMillis();
        m.go();
        System.out.println("Done. Took " +
                           ((System.currentTimeMillis() - startTime)/1000.0)
                           + "s");
        System.exit(0);
    }
}
