package featextractors;

import java.util.ArrayList;
import javax.sound.sampled.*;
import java.io.*;

/**
 * A class to extract and hold STFT info from an AudioInputStream.
 * Calls listeners when a new frame is added.  Calculates the
 * log magnitude spectrogram.
 *
 * @author Mike Mandel (mim@ee.columbia.edu)
 */
public class STFT {
  AudioInputStream input;
  AudioFormat format;
  int bytesPerWavFrame, frameLen;
  ArrayList listeners = new ArrayList();
  double[] re, im, window;
  double log10 = Math.log(10);
  double epsilon = 1e-9;       // avoid log of zero
  RingMatrix freq, time;
  FFT fft;

  double rmsTarget = 0.08;
  double rmsAlpha = 0.001;
  double rms = 1;

  float samplingRate = 44100;  

  // The line should be open, but not started yet.
  public STFT(AudioInputStream input, int frameLen, int history) {
    freq = new RingMatrix(frameLen/2+1, history);
    time = new RingMatrix(frameLen, history);
    this.frameLen = frameLen;

    this.input = input;
    format = input.getFormat();
    bytesPerWavFrame = format.getFrameSize();

    samplingRate = format.getSampleRate();

    fft = new FFT(frameLen);

    this.re = new double[frameLen];
    this.im = new double[frameLen];
    this.window = fft.getWindow();
    for(int i=0; i<im.length; i++)
      im[i] = 0;
  }

  public void start() {
    byte[] b = new byte[bytesPerWavFrame * frameLen];

    int bytesRead = 22;
    while(bytesRead > 0) {
      try {
	bytesRead = input.read(b);
      } catch(IOException ioe) {
	ioe.printStackTrace();
	return;
      }
//       System.out.println(bytesRead + " bytes read");
//       System.out.print(".");

      // store the unwindowed waveform for getSamples function
      double[] wav = time.checkOutColumn();
      bytes2doubles(b, wav);

      // Normalize rms using a moving average estimate of it
      // Calculate current rms
      double rmsCur = 0;
      for(int i=0; i<wav.length; i++)
	rmsCur += wav[i]*wav[i];
      rmsCur = Math.sqrt(rmsCur / wav.length);

      // update moving average
      rms = rmsAlpha*rmsCur + (1-rmsAlpha)*rms;

      // normalize by rms
      for(int i=0; i<wav.length; i++)
	wav[i] = wav[i] * rmsTarget / rms;

      time.checkInColumn(wav);
	
      // window waveform
      for(int i=0; i<wav.length; i++)
	re[i] = window[i] * wav[i];
	
      // take fft
      fft.fft(re, im);
      
      // Calculate magnitude
      double[] mag = freq.checkOutColumn();
      for(int i=0; i<mag.length; i++)
// 	mag[i] = re[i]*re[i] + im[i]*im[i];
	mag[i] = 10*Math.log(re[i]*re[i] + im[i]*im[i] + epsilon) / log10;

      // clear im[]
      for(int i=0; i<im.length; i++)
	im[i] = 0;
	
      // Tell everyone concerned that we've added another frame
      long frAddr = freq.checkInColumn(mag);
      notifyListeners(frAddr);
    }
  }


  public static void do_stft(AudioInputStream input)
  {
  }

  // Get the waveform samples from frames frStart to frEnd-1
  public double[] getSamples(long frStart, long frEnd) {
    long sampStart = fr2Samp(frStart);
    long sampEnd = fr2Samp(frEnd);

//     System.out.println("sampStart=" + sampStart + ", sampEnd=" + sampEnd);
//     System.out.println("frStart=" + frStart + ", frEnd=" + frEnd);

    double[] x = new double[(int)(sampEnd - sampStart)];

    for(int fr=0; fr < frEnd-frStart; fr++) {
      double[] frame = time.getColumn(frStart+fr);
      if(frame == null) continue;
      for(int i=0; i<frameLen; i++)
	x[(int)(fr2Samp(fr+frStart)-fr2Samp(frStart) + i)] = frame[i];
    }

    return x;
  }
  
  // Convert an address in frames into an address in samples
  public long fr2Samp(long frAddr) {
    // frames do not overlap, so calculation is easy
    return frameLen * frAddr;
  }

  // Convert an address in samples into an address in frames
  public long samp2fr(long sampAddr) {
    return sampAddr/frameLen;
  }

  public double[] getFrame(long frAddr) { return freq.getColumn(frAddr); }
  public int getColumns() { return freq.getColumns(); }
  public int getRows() { return freq.getRows(); }

  // Dealing with FrameListeners
  public void addFrameListener(FrameListener fl) {
    listeners.add(fl);
  }
  public void removeFrameListener(FrameListener fl) {
    listeners.remove(fl);
  }
  public void notifyListeners(long frAddr) {
    for(int i=0; i<listeners.size(); i++) {
      FrameListener list = (FrameListener) listeners.get(i);
      list.newFrame(this, frAddr);
    }
  }


  // Convert a byte stream into a stream of doubles.  If it's stereo,
  // the channels will be interleaved with each other in the double
  // stream, as in the byte stream.
  public void bytes2doubles(byte[] audioBytes, double[] audioData) {
    if (format.getSampleSizeInBits() == 16) {
      if (format.isBigEndian()) {
	for (int i = 0; i < audioData.length; i++) {
	  /* First byte is MSB (high order) */
	  int MSB = (int) audioBytes[2*i];
	  /* Second byte is LSB (low order) */
	  int LSB = (int) audioBytes[2*i+1];
	  audioData[i] = ((double)(MSB << 8 | (255 & LSB))) 
	    / 32768.0;
	}
      } else {
	for (int i = 0; i < audioData.length; i++) {
	  /* First byte is LSB (low order) */
	  int LSB = (int) audioBytes[2*i];
	  /* Second byte is MSB (high order) */
	  int MSB = (int) audioBytes[2*i+1];
	  audioData[i] = ((double)(MSB << 8 | (255 & LSB))) 
	    / 32768.0;
	}
      }
    } else if (format.getSampleSizeInBits() == 8) {
      int nlengthInSamples = audioBytes.length;
      if (format.getEncoding().toString().startsWith("PCM_SIGN")) {
	for (int i = 0; i < audioBytes.length; i++) {
	  audioData[i] = audioBytes[i] / 128.0;
	}
      } else {
	for (int i = 0; i < audioBytes.length; i++) {
	  audioData[i] = (audioBytes[i] - 128) / 128.0;
	}
      }
    }
  }

  // Convert an address in frames to an address in seconds
  public double fr2Seconds(long frAddr)
  {
      return(fr2Samp(frAddr)/samplingRate);
  }  

  // Convert an address in seconds to an address in frames
  public long seconds2fr(double sec)
  {
      return(samp2fr((long)(sec*samplingRate)));
  }  
}
